/**************************************************************************************************
	$Header: /pub/cvsroot/mydns/src/mydns/encode.c,v 1.7 2002/07/29 04:42:50 bboy Exp $
	encode.c: Encode/decode DNS-encoded names.

	Copyright (C) 2002  Don Moore <bboy@bboy.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at Your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************************************************************************/

#include "named.h"


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	NAME_REMEMBER
	Adds the specified name + offset to the `Labels' array within the specified task.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
inline void
name_remember(t, name, offset)
	TASK *t;
	u_char *name;
	u_int offset;
{
	if (!name || strlen(name) > 64)						/* Don't store labels > 64 bytes in length */
		return;

	if (!t->Names)
	{
		if (!(t->Names = malloc(sizeof(u_char *))))
			Err("malloc");
	}
	else
	{
		if (!(t->Names = realloc(t->Names, (t->numNames + 1) * sizeof(u_char *))))
			Err("realloc");
	}

	if (!(t->Names[t->numNames] = strdup(name)))
		Err("strdup");

	if (!t->Offsets)
	{
		if (!(t->Offsets = malloc(sizeof(u_int))))
			Err("malloc");
	}
	else
	{
		if (!(t->Offsets = realloc(t->Offsets, (t->numNames + 1) * sizeof(u_int))))
			Err("realloc");
	}

	t->Offsets[t->numNames] = offset;
	t->numNames++;
}
/*--- name_remember() ---------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	NAME_FIND
	Searches the task's remembered names arary for `name'.
	Returns the offset within the reply if found, or 0 if not found.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
u_int
name_find(t, name)
	TASK *t;
	u_char *name;
{
	register u_int n;

	for (n = 0; n < t->numNames; n++)
		if (!strcasecmp(t->Names[n], name))
			return (t->Offsets[n]);
	return (0);
}
/*--- name_find() -------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	NAME_UNENCODE
	Get a name from a compressed question.
	Returns the new location of src, or NULL if an error occurred.
	If an error occurs, puts a descriptive message in `dest'.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
u_char *
name_unencode(u_char *start, size_t slen, u_char *current, u_char *dest, size_t destsize)
{
	register u_char *s = current,								/* Current pointer into input */
						 *d = dest,									/* Current pointer into destination */
						 n, len, *rp = NULL;

#define CHECK_DEST_SPACE(n)  \
	if (d >= dest+(destsize-(n))) \
	{ \
		strncpy(dest, _("input too large for buffer"), destsize-1); \
		return (NULL); \
	} \

	if (*s == 0)													/* The name is just "." */
	{
		CHECK_DEST_SPACE(2);
		*d++ = '.';
		*d = '\0';
		return (s+1);
	}

	while ((len = *s))											/* Get length octet */
	{
		register u_char mask = len & 0xC0;
		if (mask == 0xC0)
		{
			u_int16_t offset;

			DNS_GET16(offset, s);
			offset &= ~0xC000;
			if (!rp)
				rp = s;
			s = start + offset;
		}
		else if (mask == 0x40 || mask == 0x80)
		{
			strncpy(dest, _("unknown/invalid compression method"), destsize-1);
			return (NULL);
		}
		else
		{
			s++;
			if (len + (s - current) > DNS_MAXNAMELEN)
			{
				strncpy(dest, _("name greater than maximum name length"), destsize-1);
				return (NULL);
			}
			for (n = 0; n < len; n++)							/* Get label */
			{
				CHECK_DEST_SPACE(1);
				*d++ = tolower(*s++);
			}
			CHECK_DEST_SPACE(1);
			*d++ = '.';
		}
	}
	CHECK_DEST_SPACE(1);
	*d = '\0';
	return (rp ? rp : s+1);
}
/*--- name_unencode() ---------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	NAME_ENCODE
	Encodes `in_name' into `dest'.  Returns the length of data in `dest', or -1 on error.
	If `name' is not NULL, it should be DNS_MAXNAMELEN bytes or bigger.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int
name_encode(TASK *t, u_char *dest, u_char *name, u_int dest_offset)
{
	register u_char *namebuf, *c, *d, *this_name, *cp;
	register int len = 0;
#ifndef WITHOUT_ENCODING
	register u_int offset;
#endif

	/* Examine name one label at a time */
	if (!(namebuf = strdup(name)))
		Err("strdup");
	for (c = namebuf, d = dest; *c; c++)
		if (c == namebuf || *c == '.')
		{
			if (!c[1])
			{
				len++;
				if (len > DNS_MAXNAMELEN)
				{
					dnserror(t, DNS_R_FORMERR);
					free(namebuf);
					return (-1);
				}
				*d++ = 0;
				free(namebuf);
				return (len);
			}
			this_name = (c == namebuf) ? c : (++c);
#ifndef WITHOUT_ENCODING
			if ((offset = name_find(t, this_name)))
			{
				/* Found marker for this name - output offset pointer and we're done */
				len += SIZE16;
				if (len > DNS_MAXNAMELEN)
				{
					dnserror(t, DNS_R_FORMERR);
					free(namebuf);
					return (-1);
				}
				offset |= 0xC000;
				DNS_PUT16(d, offset);
				free(namebuf);
				return (len);
			}
			else		/* No marker for this name; encode current label and store marker */
#endif
			{
				register u_int nlen;

				if ((cp = strchr(this_name, '.')))
					*cp = '\0';
				nlen = strlen(this_name);
				len += nlen + 1;
				if (len > DNS_MAXNAMELEN)
				{
					dnserror(t, DNS_R_FORMERR);
					free(namebuf);
					return (-1);
				}
				*d++ = (u_char)nlen;
				memcpy(d, this_name, nlen);
				d += nlen;
				if (cp)
					*cp = '.';
#ifndef WITHOUT_ENCODING
				name_remember(t, this_name, dest_offset + (c - namebuf));
#endif
			}
		}
	free(namebuf);
	return (len);
}
/*--- name_encode() -----------------------------------------------------------------------------*/

/* vi:set ts=3: */
/* NEED_PO */
